/*
Contains interface to the National Instruments mx library.
	classes : 
		CNICard : Represents card capabilities.
		CTask : status of Task currently running on card 
	NIPoller() : function which runs on own thread and polls device.

	todo - query card settings ( 8.0 required ? ) 
		int iLen;
		char buffer[1000];
		int32 iErr = DAQmxGetDevAIPhysicalChans("Dev1", buffer, 1000);
			// buffer = "Dev1/ai0, Dev1/ai1, D.....31"
		iErr = DAQmxGetDevAOPhysicalChans("Dev1", buffer, 1000);
			// buffer = "Dev1/ao0, Dev1/ao1, Dev1/ao2, Dev1/ao3"

//---- Thread routine for digital pulse output ( manually controlled via buttons, not protocol based ) 
// lpParam is ptr to BYTES : 
// [0] = 8/32 : bytes to write (i.e. physical size of port0 = 8 or 32 )
// [1] = nonzero : when thread starts, immediately send an output. If 0, wait for 1st change.
// [2] = nonzero : cancel.  Main thread sets this when it wants the poller to exit.
// [3] = ? 
// [4-7] = data; Main thread changes this and poller writes it out.
// [8-11] = last status written by the poller.  
// [12-...] = null terminated string with port names.  Ex: "Dev1/port0/line0:3"
*/
//TODO - consider runtime load dll - avoid no NI loaded issue.

//-- Acquisition of data callback routine
typedef int (__cdecl * AcqCallBack)( int iSamples, void * pAnaData, void * pDigData, void * pSession );

//-- Generation of data callback routine
typedef int (__cdecl * GenCallBack)( int iSamples, void * pData, void * pDigData, void * pSession, int iSetPos );

//-- Termination of acq/gen is complete 
typedef int (__cdecl * TermCallBack)( void * pSession );

const int MAXDIGLINES = 32;

//----- Feedback class as a base for specific feedback calculations.  Initially just PID controller.
const int FEEDBACKTYPE_PID = 1;
class CFeedback : public CItem { 
	public : 
		int m_iFeedbackType;
		CString m_cName;
		virtual double CalculateOutput( double dSetpoint, double dInput, int iSample )=0;

		int m_iInputIndex;		// index to acq channel being compared
		int m_iOutputIndex;		// index to output channel being compared

		CFeedback * NewCopy();
		static bool LoadFromFile( CFeedback ** ppFeedback, FILE * h, int iVer );	
		bool SaveToFile( FILE * h );
		int SaveSize();
		
		virtual bool LoadHelper( FILE * h, int iVer )=0;
		virtual bool SaveHelper( FILE * h )=0;
		virtual int SaveSizeHelper()=0;
	};

//-- Implement a PID control. e=AO-AI; generate = Kp*e + Ki*sum(e[...])/dt + Kd*(de/dt)
// eventually implement as feedback class with virtuals and CPID inherits from it.
class CPID : public CFeedback { 
	public : 
		static const int PID_QUE_RECALC_INTERVAL = 1000;
		static const int PID_QUE_DEFAULT_SIZE = 10;
		CPID();
		~CPID();
		
		void Set( double dDeltaT, double dKp, double dKi, double dKd, int iQueueSize, int iAIIndex, int iAOIndex );
		double m_dDeltaT;		// Time between samples 
		double m_dKp;			// constant for proportional component
		double m_dKi;			// constant for integeral component
		double m_dKd;			// constant for differential component
		
		double CalculateOutput( double dSetpoint, double dInput, int iSample );
		
		//-- maintain queue of calls to CalculateOutput for integral calculation
		int m_iQueueSize;		// # of entries in Queue
		int m_iQueueLast;		// index to last position in queue filled or -1 for uninitialized queue 
		double * m_adQueueValue;// data queue
		int * m_aiQueueTime;	// Time Queue 
		int m_iQueueSummed;		// Calls to CalculateOutput since last QueueClear()
		bool QueueAllocate( int iQueueSize );
		double m_dQueueSum;

		//-- persistance
		bool LoadHelper( FILE * h, int iVer );
		bool SaveHelper( FILE * h );
		int SaveSizeHelper();
	};

//-- abstraction of National Instrument Card
class CNICard { 
	public : 
		const static int NI_NAMESIZE = 64;		// Size of stored card names (Dev1) and model numbers (PCI 6229)
		const static int NI_GAINS = 4;

	//----- Card properties 
	public : 
		int m_iCard;
		char m_achName[NI_NAMESIZE];
		double m_dInRange;			// maximum input voltage ( ex. 10.0 = +/-10 volts )
		double m_dOutRange;			// maximum output voltage ( ex. 10.0 = +/-10 volts )

		int m_iAIChannels;			// Count of analog input channels
		int m_iAIBits;				// analog input resolution bits
		int m_iMaxIRate;			// Maximum hz. input rate 
		double m_afIGain[NI_GAINS];	// Array of allowable gain settings
		int m_iIGains;				// Number of gains in array m_aiIGain

		int m_iAOChannels;			// Count of analog output channels
		int m_iAOBits;				// analog output resolution bits
		int m_iMaxORate;			// Maximum hz. output rate 
		double m_afOGain[NI_GAINS];	// Array of allowable gain settings
		int m_iOGains;				// Number of gains in array m_aiOGain

		int m_iDChannels;			// number of digital I/O channels
		
		// For provided card #, set all the above member vars.  
		// Return true if card is recognized.
		bool Set( int iCard, const char * pchName, const char * pchModel  );	
	
		bool m_bValid;				// true if card is a known NI card # 
		char m_achModel[NI_NAMESIZE];
		};

struct NIPOLLER_INIT;

class CTask { 
	public : 
		CNICard * m_poCard;			// pointer to card 
		int32 m_iError;				// shared error code 
		HANDLE m_hPoller;			// Handle to thread Polling current acq/gen

		//-- Acquisition		
		int m_iAcqSampling;			// Input sampling rate per second
		int m_iAcqBufferSize;

		TaskHandle m_hAnaAcq;			// handle for analog acquisition task
		bool m_bAnaAcqTask;			// true if input task is being created
		int m_iAcqChannels;			// count of input channels in task

		TaskHandle m_hDigAcq;		// handle for digital acquisition task
		bool m_bDigAcqTask;
		int m_iDigAcqLines;

		//-- Generation 
		int m_iGenSampling;			// Output sampling rate 
		int m_iGenBufferSize;		
		bool m_bGenFinite;				// Output only task has finite size 

		TaskHandle m_hAnaGen;			// handle for analog generation task
		bool m_bAnaGenTask;			// true if output task is being created
		int m_iGenChannels;			// count of output channels in task

		TaskHandle m_hDigGen;		// handle for digital generation task
		bool m_bDigGenTask;
		int m_iDigGenLines;
		bool m_b100khz;				// a dig gen only task was started, so the 100khz clock is in use
		
	public :
		// create gen and/or acq tasks 
		CTask( CNICard * poCard, 
				bool bAcq, int iAcqSampling, int iAcqBufferSize, 
				bool bGen, int iGenSampling, int iGenBufferSize, 
				bool bFeedback, bool bDigAcq, bool bDigGen );

		// add channels to the tasks
		bool AddInputChannel( int iCh, double dGain, int iMode );
		bool AddOutputChannel( int iCh, double dGain=1.0 );
		bool AddDigitalInputChannels( int iPort, int iLines, BYTE * aiLines );
		bool AddDigitalOutputChannels( int iPort, int iLines, BYTE * aiLines );

	protected : 		
		// create a Polling thread and set up timing then start the acq
		bool SetRegenMode( TaskHandle h, bool bAllowRegen );		//-- Prevent or allow glitching
		bool SetOutputBuffer( TaskHandle h, int iSize );
		bool SetInputBuffer( TaskHandle h, int iSize );
		bool SetTiming( TaskHandle h, int iRate, int iSize, char * pchTimingSource=NULL, bool bGenFinite=false );
		bool PreloadGenData( TaskHandle hAna, TaskHandle hDig, GenCallBack pFn, void * pSession );
		
	public :
		bool StartTasks( NIPOLLER_INIT * pInit, bool bGenFinite );
		};	

//-- poller communiocation mechanisms with Session, Task
struct NIPOLLER_INIT {					// m_pfn...CallBack = pointer to callback functions 
	AcqCallBack m_pfnAcqCallBack;		// Process acquired analog data 
	GenCallBack m_pfnGenCallBack;		// Fill analog generation buffer
	TermCallBack m_pfnTerminate;		// Function to call to notify task that termination has occurred (abort or complete)
	CTask * m_pTask;					// pointer to task object to communicate with 
	bool m_bCancelled;					// boolean to be set by task to notify poller of desire to terminate 
	void * m_pSession;					// pointer to initiating session 
	};

NIPOLLER_INIT * CreateInit( 
		AcqCallBack pfnAcqCallBack, 
		GenCallBack pfnGenCallBack,
		TermCallBack pfnTermCallBack, void * pSession );

//-- Thread function for NI poller.	
DWORD WINAPI NIPoller( LPVOID pInit );
	
int NiCardCount();				// Number of cards discovered, -1 = not yet initialized.
CNICard * NiCard(int iCard);	// Return pointer to specified card ( or null ) 
int NiCardFind( const char * pchName );
								// Return index to card of specified name
int NIError(int iCode=-1);		// Get / Set the error code for the last NI function call

bool NIInit();					// Initialize the NI card list.  Called once per app instance.

//-- Digital output continuously polling as background thread
DWORD WINAPI DigitalOutputPoller( LPVOID lpParam );

//-- Single digital output 
//int DigOut32( int iVal, char * pchPorts );


